from oscar.core.loading import get_model

from ..cache import (
    BillingAddressCache,
    EmailAddressCache,
    ShippingAddressCache,
    ShippingMethodCache,
)
from .base import BaseTest

Country = get_model("address", "Country")


class EmailAddressCacheTest(BaseTest):
    def test_cache(self):
        EmailAddressCache(1).set({"email": "foo1@example.com"})
        EmailAddressCache(2).set({"email": "foo2@example.com"})
        self.assertEqual(EmailAddressCache(1).get(), {"email": "foo1@example.com"})
        self.assertEqual(EmailAddressCache(2).get(), {"email": "foo2@example.com"})


class ShippingAddressCacheTest(BaseTest):
    def test_cache(self):
        ShippingAddressCache(1).set(
            {
                "first_name": "Bart",
                "last_name": "Simpson",
                "line1": "123 Evergreen Terrace",
                "line4": "Springfield",
                "state": "NY",
                "postcode": "10001",
                "country": Country.objects.get(iso_3166_1_a3="USA"),
            }
        )
        self.assertEqual(
            ShippingAddressCache(1).get(),
            {
                "first_name": "Bart",
                "last_name": "Simpson",
                "line1": "123 Evergreen Terrace",
                "line4": "Springfield",
                "state": "NY",
                "postcode": "10001",
                "country": Country.objects.get(iso_3166_1_a3="USA"),
            },
        )


class BillingAddressCacheTest(BaseTest):
    def test_cache(self):
        BillingAddressCache(1).set(
            {
                "first_name": "Lisa",
                "last_name": "Simpson",
                "line1": "123 Evergreen Terrace",
                "line4": "Springfield",
                "state": "NY",
                "postcode": "10001",
                "country": Country.objects.get(iso_3166_1_a3="USA"),
            }
        )
        self.assertEqual(
            BillingAddressCache(1).get(),
            {
                "first_name": "Lisa",
                "last_name": "Simpson",
                "line1": "123 Evergreen Terrace",
                "line4": "Springfield",
                "state": "NY",
                "postcode": "10001",
                "country": Country.objects.get(iso_3166_1_a3="USA"),
            },
        )


class ShippingMethodCacheTest(BaseTest):
    def test_cache(self):
        ShippingMethodCache(1).set(
            {
                "code": "ups-next-day-air",
                "name": "UPS Next Day Air",
                "price": "47.00",
            }
        )
        self.assertEqual(
            ShippingMethodCache(1).get(),
            {
                "code": "ups-next-day-air",
                "name": "UPS Next Day Air",
                "price": "47.00",
            },
        )



from django.contrib.auth.models import User
from django.http import HttpRequest


class PaymentMethodPermission:
    def is_permitted(
        self,
        request: HttpRequest | None = None,
        user: User | None = None,
    ) -> bool:
        raise NotImplementedError("Class must implement is_method_permitted(request=None, user=None)")


class Public(PaymentMethodPermission):
    def is_permitted(
        self,
        request: HttpRequest | None = None,
        user: User | None = None,
    ) -> bool:
        return True


class StaffOnly(PaymentMethodPermission):
    def is_permitted(
        self,
        request: HttpRequest | None = None,
        user: User | None = None,
    ) -> bool:
        return user is not None and user.is_authenticated and user.is_staff


class CustomerOnly(PaymentMethodPermission):
    def is_permitted(
        self,
        request: HttpRequest | None = None,
        user: User | None = None,
    ) -> bool:
        return user is None or not user.is_authenticated or (user.is_authenticated and not user.is_staff)



from rest_framework import status
from rest_framework.reverse import reverse

from .base import BaseTest


class PaymentMethodsViewTest(BaseTest):
    maxDiff = None

    def test_introspection_authenticated(self):
        self.login(is_staff=True)
        url = reverse("api-checkout-payment-methods")
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            resp.data,
            {
                "credit-card": {
                    "type": "nested object",
                    "required": False,
                    "read_only": False,
                    "label": "Credit Card",
                    "children": {
                        "method_type": {
                            "type": "choice",
                            "required": True,
                            "read_only": False,
                            "label": "Method type",
                            "choices": [{"value": "credit-card", "display_name": "Credit Card"}],
                        },
                        "enabled": {
                            "type": "boolean",
                            "required": False,
                            "read_only": False,
                            "label": "Enabled",
                        },
                        "pay_balance": {
                            "type": "boolean",
                            "required": False,
                            "read_only": False,
                            "label": "Pay balance",
                        },
                        "amount": {
                            "type": "decimal",
                            "required": False,
                            "read_only": False,
                            "label": "Amount",
                            "decimal_places": 2,
                            "max_digits": 12,
                        },
                        "reference": {
                            "type": "string",
                            "required": False,
                            "read_only": False,
                            "label": "Reference",
                            "max_length": 128,
                        },
                    },
                },
                "client-side-card": {
                    "type": "nested object",
                    "required": False,
                    "read_only": False,
                    "label": "Client-Side Card",
                    "children": {
                        "method_type": {
                            "type": "choice",
                            "required": True,
                            "read_only": False,
                            "label": "Method type",
                            "choices": [{"value": "client-side-card", "display_name": "Client-Side Card"}],
                        },
                        "enabled": {
                            "type": "boolean",
                            "required": False,
                            "read_only": False,
                            "label": "Enabled",
                        },
                        "pay_balance": {
                            "type": "boolean",
                            "required": False,
                            "read_only": False,
                            "label": "Pay balance",
                        },
                        "amount": {
                            "type": "decimal",
                            "required": False,
                            "read_only": False,
                            "label": "Amount",
                            "decimal_places": 2,
                            "max_digits": 12,
                        },
                        "reference": {
                            "type": "string",
                            "required": False,
                            "read_only": False,
                            "label": "Reference",
                            "max_length": 128,
                        },
                    },
                },
                "cash": {
                    "type": "nested object",
                    "required": False,
                    "read_only": False,
                    "label": "Cash",
                    "children": {
                        "method_type": {
                            "type": "choice",
                            "required": True,
                            "read_only": False,
                            "label": "Method type",
                            "choices": [{"value": "cash", "display_name": "Cash"}],
                        },
                        "enabled": {
                            "type": "boolean",
                            "required": False,
                            "read_only": False,
                            "label": "Enabled",
                        },
                        "pay_balance": {
                            "type": "boolean",
                            "required": False,
                            "read_only": False,
                            "label": "Pay balance",
                        },
                        "amount": {
                            "type": "decimal",
                            "required": False,
                            "read_only": False,
                            "label": "Amount",
                            "decimal_places": 2,
                            "max_digits": 12,
                        },
                        "reference": {
                            "type": "string",
                            "required": False,
                            "read_only": False,
                            "label": "Reference",
                            "max_length": 128,
                        },
                    },
                },
                "pay-later": {
                    "type": "nested object",
                    "required": False,
                    "read_only": False,
                    "label": "Pay Later",
                    "children": {
                        "method_type": {
                            "type": "choice",
                            "required": True,
                            "read_only": False,
                            "label": "Method type",
                            "choices": [{"value": "pay-later", "display_name": "Pay Later"}],
                        },
                        "enabled": {
                            "type": "boolean",
                            "required": False,
                            "read_only": False,
                            "label": "Enabled",
                        },
                        "pay_balance": {
                            "type": "boolean",
                            "required": False,
                            "read_only": False,
                            "label": "Pay balance",
                        },
                        "amount": {
                            "type": "decimal",
                            "required": False,
                            "read_only": False,
                            "label": "Amount",
                            "decimal_places": 2,
                            "max_digits": 12,
                        },
                        "reference": {
                            "type": "string",
                            "required": False,
                            "read_only": False,
                            "label": "Reference",
                            "max_length": 128,
                        },
                    },
                },
            },
        )

    def test_introspection_anonymous(self):
        url = reverse("api-checkout-payment-methods")
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            resp.data,
            {
                "credit-card": {
                    "type": "nested object",
                    "required": False,
                    "read_only": False,
                    "label": "Credit Card",
                    "children": {
                        "method_type": {
                            "type": "choice",
                            "required": True,
                            "read_only": False,
                            "label": "Method type",
                            "choices": [{"value": "credit-card", "display_name": "Credit Card"}],
                        },
                        "enabled": {
                            "type": "boolean",
                            "required": False,
                            "read_only": False,
                            "label": "Enabled",
                        },
                        "pay_balance": {
                            "type": "boolean",
                            "required": False,
                            "read_only": False,
                            "label": "Pay balance",
                        },
                        "amount": {
                            "type": "decimal",
                            "required": False,
                            "read_only": False,
                            "label": "Amount",
                            "decimal_places": 2,
                            "max_digits": 12,
                        },
                        "reference": {
                            "type": "string",
                            "required": False,
                            "read_only": False,
                            "label": "Reference",
                            "max_length": 128,
                        },
                    },
                },
                "client-side-card": {
                    "type": "nested object",
                    "required": False,
                    "read_only": False,
                    "label": "Client-Side Card",
                    "children": {
                        "method_type": {
                            "type": "choice",
                            "required": True,
                            "read_only": False,
                            "label": "Method type",
                            "choices": [{"value": "client-side-card", "display_name": "Client-Side Card"}],
                        },
                        "enabled": {
                            "type": "boolean",
                            "required": False,
                            "read_only": False,
                            "label": "Enabled",
                        },
                        "pay_balance": {
                            "type": "boolean",
                            "required": False,
                            "read_only": False,
                            "label": "Pay balance",
                        },
                        "amount": {
                            "type": "decimal",
                            "required": False,
                            "read_only": False,
                            "label": "Amount",
                            "decimal_places": 2,
                            "max_digits": 12,
                        },
                        "reference": {
                            "type": "string",
                            "required": False,
                            "read_only": False,
                            "label": "Reference",
                            "max_length": 128,
                        },
                    },
                },
                "pay-later": {
                    "type": "nested object",
                    "required": False,
                    "read_only": False,
                    "label": "Pay Later",
                    "children": {
                        "method_type": {
                            "type": "choice",
                            "required": True,
                            "read_only": False,
                            "label": "Method type",
                            "choices": [{"value": "pay-later", "display_name": "Pay Later"}],
                        },
                        "enabled": {
                            "type": "boolean",
                            "required": False,
                            "read_only": False,
                            "label": "Enabled",
                        },
                        "pay_balance": {
                            "type": "boolean",
                            "required": False,
                            "read_only": False,
                            "label": "Pay balance",
                        },
                        "amount": {
                            "type": "decimal",
                            "required": False,
                            "read_only": False,
                            "label": "Amount",
                            "decimal_places": 2,
                            "max_digits": 12,
                        },
                        "reference": {
                            "type": "string",
                            "required": False,
                            "read_only": False,
                            "label": "Reference",
                            "max_length": 128,
                        },
                    },
                },
            },
        )



from rest_framework.test import APIRequestFactory

from ..permissions import Public, StaffOnly
from .base import BaseTest


class PublicTest(BaseTest):
    def test_public_anon(self):
        factory = APIRequestFactory()
        request = factory.get("/")
        self.assertTrue(Public().is_permitted(request))

    def test_public_not_staff(self):
        user = self.login(is_staff=False)
        factory = APIRequestFactory()
        request = factory.get("/")
        self.assertTrue(Public().is_permitted(request, user))

    def test_public_is_staff(self):
        user = self.login(is_staff=True)
        factory = APIRequestFactory()
        request = factory.get("/")
        self.assertTrue(Public().is_permitted(request, user))


class StaffOnlyTest(BaseTest):
    def test_staff_only_anon(self):
        factory = APIRequestFactory()
        request = factory.get("/")
        self.assertFalse(StaffOnly().is_permitted(request))

    def test_staff_only_not_staff(self):
        user = self.login(is_staff=False)
        factory = APIRequestFactory()
        request = factory.get("/")
        self.assertFalse(StaffOnly().is_permitted(request, user))

    def test_staff_only_is_staff(self):
        user = self.login(is_staff=True)
        factory = APIRequestFactory()
        request = factory.get("/")
        self.assertTrue(StaffOnly().is_permitted(request, user))



from decimal import Decimal
from typing import TYPE_CHECKING, Any

from django.core.signing import Signer
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _
from oscar.apps.payment.models import Transaction
from oscarapicheckout.methods import (
    PaymentMethod,
    PaymentMethodData,
    PaymentMethodSerializer,
)
from oscarapicheckout.states import (
    ClientSidePaymentRequired,
    Complete,
    Declined,
)

if TYPE_CHECKING:
    from ..order.models import Order


class ClientSideCard(PaymentMethod[PaymentMethodData]):
    """
    Example of a payment method that uses a client-side SDK (e.g. CyberSource
    Unified Checkout, Stripe Elements, Braintree Drop-in, Adyen Web Components).

    Instead of returning a form-post URL, it returns an opaque token that the
    client uses to initialize a provider-hosted JS widget.
    """

    name = _("Client-Side Card")
    code = "client-side-card"
    serializer_class = PaymentMethodSerializer[PaymentMethodData]

    def _record_payment(
        self,
        request: HttpRequest,
        order: "Order",
        method_key: str,
        amount: Decimal,
        reference: str,
        **kwargs: Any,
    ) -> ClientSidePaymentRequired:
        return ClientSidePaymentRequired(
            amount=amount,
            payment_processor="sandbox-processor",
            data={
                "amount": str(amount),
                "token": "sandbox-client-token-xyz",
                "sdk_url": "https://sandbox.example.com/sdk.js",
                "reference_number": order.number,
                "transaction_id": Signer().sign(method_key),
            },
        )

    def record_successful_authorization(
        self,
        order: "Order",
        amount: Decimal,
        reference: str,
    ) -> Complete:
        source = self.get_source(order, reference)

        source.allocate(amount, reference=reference, status="ACCEPTED")
        event = self.make_authorize_event(order, amount, reference)
        for line in order.lines.all():
            self.make_event_quantity(event, line, line.quantity)

        return Complete(amount, source_id=source.pk)

    def record_declined_authorization(
        self,
        order: "Order",
        amount: Decimal,
        reference: str,
    ) -> Declined:
        source = self.get_source(order, reference)
        source._create_transaction(Transaction.AUTHORISE, amount, reference=reference, status="DECLINED")
        return Declined(amount, source_id=source.pk)



from collections.abc import Sequence
from decimal import Decimal
from enum import UNIQUE, StrEnum, verify
from typing import Any, Literal, TypedDict


@verify(UNIQUE)
class PaymentMethodStatus(StrEnum):
    PENDING = "Pending"
    DECLINED = "Declined"
    COMPLETE = "Complete"
    DEFERRED = "Deferred"
    CONSUMED = "Consumed"


# For backwards compat
PENDING = PaymentMethodStatus.PENDING
DECLINED = PaymentMethodStatus.DECLINED
COMPLETE = PaymentMethodStatus.COMPLETE
DEFERRED = PaymentMethodStatus.DEFERRED
CONSUMED = PaymentMethodStatus.CONSUMED


class FormPostRequiredFormDataField(TypedDict):
    key: str
    value: str


class FormPostRequiredFormData(TypedDict):
    type: Literal["form"]
    name: str
    url: str
    method: str
    fields: Sequence[FormPostRequiredFormDataField]


class ClientSidePaymentData(TypedDict):
    type: Literal["client-side-payment"]
    payment_processor: str
    data: dict[str, Any]


type RequiredAction = FormPostRequiredFormData | ClientSidePaymentData | None


class PaymentStatus:
    status: PaymentMethodStatus
    amount: Decimal

    def __init__(self, amount: Decimal) -> None:
        self.amount = amount

    def get_required_action(self) -> RequiredAction:
        raise NotImplementedError("Subclass does not implement get_required_action()")


class SourceBoundPaymentStatus(PaymentStatus):
    def __init__(
        self,
        amount: Decimal,
        source_id: int | None = None,
    ) -> None:
        super().__init__(amount)
        self.source_id = source_id


class Complete(SourceBoundPaymentStatus):
    status = PaymentMethodStatus.COMPLETE


class Deferred(SourceBoundPaymentStatus):
    status = PaymentMethodStatus.DEFERRED


class Declined(SourceBoundPaymentStatus):
    status = PaymentMethodStatus.DECLINED


class Consumed(SourceBoundPaymentStatus):
    status = PaymentMethodStatus.CONSUMED


class FormPostRequired(PaymentStatus):
    status = PaymentMethodStatus.PENDING
    form_data: FormPostRequiredFormData

    def __init__(
        self,
        amount: Decimal,
        name: str,
        url: str,
        method: str = "POST",
        fields: Sequence[FormPostRequiredFormDataField] = [],
    ) -> None:
        super().__init__(amount)
        self.form_data = FormPostRequiredFormData(
            type="form",
            name=name,
            url=url,
            method=method,
            fields=fields,
        )

    def get_required_action(self) -> RequiredAction:
        return self.form_data


class ClientSidePaymentRequired(PaymentStatus):
    status = PaymentMethodStatus.PENDING
    action_data: ClientSidePaymentData

    def __init__(
        self,
        amount: Decimal,
        payment_processor: str,
        data: dict[str, Any],
    ) -> None:
        super().__init__(amount)
        self.action_data = ClientSidePaymentData(
            type="client-side-payment",
            payment_processor=payment_processor,
            data=data,
        )

    def get_required_action(self) -> RequiredAction:
        return self.action_data



from django.db import migrations, models
import oscar.core.utils


class Migration(migrations.Migration):
    dependencies = [
        ("order", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="order",
            name="currency",
            field=models.CharField(
                default=oscar.core.utils.get_default_currency,
                max_length=12,
                verbose_name="Currency",
            ),
        ),
    ]



# Generated by Django 5.2.7 on 2025-11-19 18:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("order", "0015_alter_line_num_allocated"),
    ]

    operations = [
        migrations.AddField(
            model_name="order",
            name="analytics_tracked",
            field=models.BooleanField(default=False),
        ),
    ]



from typing import Any
import logging

from django.db import transaction
from django.dispatch import receiver
from django.http import HttpRequest
from oscar.apps.order.signals import order_status_changed
from oscar.core.loading import get_model

from .email import OrderMessageSender
from .settings import ORDER_STATUS_PAYMENT_DECLINED
from .signals import order_payment_authorized

Order = get_model("order", "Order")

logger = logging.getLogger(__name__)


@receiver(order_payment_authorized)
def send_order_confirmation_message(
    sender: type[Any],
    order: Order,
    request: HttpRequest,
    **kwargs: Any,
) -> None:
    transaction.on_commit(lambda: OrderMessageSender(request).send_order_placed_email(order))


@receiver(order_status_changed)
def update_basket_status_upon_order_status_change(
    sender: type[Any],
    order: Order,
    old_status: str,
    new_status: str,
    **kwargs: Any,
) -> None:
    """
    When an order transitions from "Payment Declined" to any other status, make sure
    it's associated basket is not still editable.
    """
    basket = order.basket
    if not basket:
        return

    if new_status == ORDER_STATUS_PAYMENT_DECLINED:
        return

    if not basket.can_be_edited:
        return

    logger.info(
        "Changing status of Basket[%s] from %s to Submitted due to order status change from %s to %s.",
        basket.pk,
        basket.status,
        old_status,
        new_status,
    )
    basket.submit()



from typing import Any

from django.shortcuts import get_object_or_404
from oscar.core.loading import get_model
from rest_framework import generics, status
from rest_framework.request import Request
from rest_framework.response import Response

from . import utils
from .methods import PaymentMethod, PaymentMethodData
from .serializers import (
    CheckoutSerializer,
    CompleteDeferredPaymentSerializer,
    OrderSerializer,
    PaymentMethodsSerializer,
    PaymentStateSerializer,
)
from .signals import order_placed
from .states import CONSUMED, DECLINED, PaymentStatus

Order = get_model("order", "Order")

CHECKOUT_ORDER_ID = "checkout_order_id"


class PaymentMethodsView(generics.GenericAPIView[Any]):
    serializer_class = PaymentMethodsSerializer  # type:ignore[assignment]

    def get(self, request: Request) -> Response:
        root_serializer: PaymentMethodsSerializer = (
            self.get_serializer()  # type:ignore[assignment]
        )
        meta = self.metadata_class()  # type:ignore[operator, misc]
        data = {}
        for (
            method_code,
            method_serializer,
        ) in root_serializer.child.type_mapping.items():
            method = root_serializer.methods[method_code]
            data[method_code] = {
                "type": "nested object",
                "required": False,
                "read_only": False,
                "label": method.name,
                "children": meta.get_serializer_info(method_serializer),
            }
        return Response(data)


class CheckoutView(generics.GenericAPIView[Any]):
    """
    Checkout and begin collecting payment.

    POST(basket, shipping_address, wfrs_source_account,
         [total, shipping_method_code, shipping_charge, billing_address]):
    {
        "basket": "/api/baskets/1/",
        "guest_email": "foo@example.com",
        "total": "100.0",
        "shipping_charge": {
            "currency": "EUR",
            "excl_tax": "10.0",
            "tax": "0.6"
        },
        "shipping_method_code": "no-shipping-required",
        "shipping_address": {
            "country": "/api/countries/NL/",
            "first_name": "Henk",
            "last_name": "Van den Heuvel",
            "line1": "Roemerlaan 44",
            "line2": "",
            "line3": "",
            "line4": "Kroekingen",
            "notes": "Niet STUK MAKEN OK!!!!",
            "phone_number": "+31 26 370 4887",
            "postcode": "7777KK",
            "state": "Gerendrecht",
            "title": "Mr"
        },
        "payment": {
            "cash": {
                "amount": "100.00"
            }
        }
    }

    Returns the order object.
    """

    serializer_class = CheckoutSerializer

    def post(self, request: Request, format: str | None = None) -> Response:
        # Wipe out any previous state data
        utils.clear_consumed_payment_method_states(request)

        # Validate the input
        c_ser: CheckoutSerializer = self.get_serializer(  # type:ignore[assignment]
            data=request.data
        )
        if not c_ser.is_valid():
            return Response(c_ser.errors, status.HTTP_406_NOT_ACCEPTABLE)

        # Freeze basket
        basket = c_ser.validated_data.get("basket")
        basket.freeze()

        # Save Order
        order = c_ser.save()
        request.session[CHECKOUT_ORDER_ID] = order.id

        # Send order_placed signal
        order_placed.send(
            sender=self,
            order=order,
            user=request.user,
            request=request,
            recaptcha_score=c_ser.get_recaptcha_score(),
        )

        # Save payment steps into session for processing
        previous_states = utils.list_payment_method_states(request)
        new_states = self._record_payments(
            previous_states=previous_states,
            request=request,
            order=order,
            methods=c_ser.fields["payment"].methods,  # type:ignore[attr-defined]
            data=c_ser.validated_data["payment"],
        )
        utils.set_payment_method_states(order, request, new_states)

        # Return order data
        o_ser = OrderSerializer(order, context={"request": request})
        return Response(o_ser.data)

    def _record_payments(
        self,
        previous_states: dict[str, PaymentStatus],
        request: Request,
        order: Order,
        methods: dict[str, PaymentMethod[PaymentMethodData]],
        data: dict[str, PaymentMethodData],
    ) -> dict[str, PaymentStatus]:
        order_balance = [order.total_incl_tax]
        new_states: dict[str, PaymentStatus] = {}

        def record(method_key: str, method_data: PaymentMethodData) -> PaymentStatus:
            # If a previous payment method at least partially succeeded, hasn't been consumed by an
            # order, and is for the same amount, recycle it. This requires that the amount hasn't changed.

            # Get the processor class for this method
            code = method_data["method_type"]
            method = methods[code]

            state: PaymentStatus | None = None
            if method_key in previous_states:
                prev = previous_states[method_key]
                if prev.status not in (DECLINED, CONSUMED):
                    if prev.amount == method_data["amount"]:
                        state = prev
                    else:
                        # Previous payment exists but we can't recycle it; void whatever already exists.
                        method.void_existing_payment(request, order, method_key, prev)

            # Previous payment method doesn't exist or can't be reused. Create it now.
            final_state = state if state else method.record_payment(request, order, method_key, **method_data)
            # Subtract amount from pending order balance.
            order_balance[0] = order_balance[0] - final_state.amount
            return final_state

        # Loop through each method with a specified amount to charge
        data_amount_specified = {k: v for k, v in data.items() if not v["pay_balance"]}
        for key, method_data in data_amount_specified.items():
            new_states[key] = record(key, method_data)

        # Change the remainder, not covered by the above methods, to the method marked with `pay_balance`
        data_pay_balance = {k: v for k, v in data.items() if v["pay_balance"]}
        for key, method_data in data_pay_balance.items():
            method_data["amount"] = order_balance[0]
            new_states[key] = record(key, method_data)

        return new_states


class CompleteDeferredPaymentView(CheckoutView):
    """
    Authorize payment for an order previously placed using the “Pay Later”
    deferred payment method.
    """

    serializer_class = CompleteDeferredPaymentSerializer  # type:ignore[assignment]

    def post(self, request: Request, format: str | None = None) -> Response:
        # Wipe out any previous state data
        utils.clear_consumed_payment_method_states(request)

        # Validate the input
        c_ser: CompleteDeferredPaymentSerializer = self.get_serializer(  # type:ignore[assignment]
            data=request.data
        )
        if not c_ser.is_valid():
            return Response(c_ser.errors, status.HTTP_406_NOT_ACCEPTABLE)

        # Update the session to note that we're working on this order
        order = c_ser.validated_data["order"]
        request.session[CHECKOUT_ORDER_ID] = order.id

        # Save payment steps into session for processing
        previous_states = utils.list_payment_method_states(request)
        new_states = self._record_payments(
            previous_states=previous_states,
            request=request,
            order=order,
            methods=c_ser.fields["payment"].methods,  # type:ignore[attr-defined]
            data=c_ser.validated_data["payment"],
        )
        utils.set_payment_method_states(order, request, new_states)

        # Return order data
        o_ser = OrderSerializer(order, context={"request": request})
        return Response(o_ser.data)


class PaymentStatesView(generics.GenericAPIView[Any]):
    def get(self, request: Request, pk: int | None = None) -> Response:
        # We don't really use the provided pk. It's just there to be compatible with oscarapi
        if pk and int(pk) != request.session.get(CHECKOUT_ORDER_ID):
            return Response(status=status.HTTP_404_NOT_FOUND)

        # Fetch the order object from the session, but only if it's pending
        pk = request.session.get(CHECKOUT_ORDER_ID)
        order = get_object_or_404(Order, pk=pk)

        # Return order status and payment states
        states = utils.list_payment_method_states(request)
        state_data = {}
        for key, state in states.items():
            ser = PaymentStateSerializer(instance=state)
            state_data[key] = ser.data

        return Response(
            {
                "order_status": order.status,
                "payment_method_states": state_data if any(state_data) else None,
            }
        )



from decimal import Decimal
from typing import TYPE_CHECKING, Any

from django.core.signing import Signer
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _
from oscar.apps.payment.models import Transaction
from oscarapicheckout.methods import (
    PaymentMethod,
    PaymentMethodData,
    PaymentMethodSerializer,
)
from oscarapicheckout.states import (
    Complete,
    Declined,
    FormPostRequired,
    FormPostRequiredFormDataField,
)
from rest_framework.reverse import reverse

if TYPE_CHECKING:
    from ..order.models import Order


class CreditCard(PaymentMethod[PaymentMethodData]):
    """
    This is an example of how to implement a payment method that required some off-site
    interaction, like Cybersource Secure Acceptance, for example. It returns a pending
    status initially that requires the client app to make a form post, which in-turn
    redirects back to us. This is a common pattern in PCI SAQ A-EP ecommerce sites.
    """

    name = _("Credit Card")
    code = "credit-card"
    serializer_class = PaymentMethodSerializer[PaymentMethodData]

    def _record_payment(
        self,
        request: HttpRequest,
        order: "Order",
        method_key: str,
        amount: Decimal,
        reference: str,
        **kwargs: Any,
    ) -> FormPostRequired:
        """Payment Step 1"""
        fields: list[FormPostRequiredFormDataField] = [
            {
                "key": "amount",
                "value": str(amount),
            },
            {
                "key": "reference_number",
                "value": order.number,
            },
            {
                "key": "transaction_id",
                "value": Signer().sign(method_key),
            },
        ]
        # Return a response showing we need to post some fields to the given
        # URL to finishing processing this payment method
        return FormPostRequired(
            amount=amount,
            name="get-token",
            url=reverse("creditcards-get-token"),
            fields=fields,
        )

    def require_authorization_post(self, order: "Order", method_key: str, amount: Decimal) -> FormPostRequired:
        """Payment Step 2"""
        fields: list[FormPostRequiredFormDataField] = [
            {
                "key": "amount",
                "value": str(amount),
            },
            {
                "key": "reference_number",
                "value": order.number,
            },
            {
                "key": "transaction_id",
                "value": Signer().sign(method_key),
            },
        ]
        return FormPostRequired(
            amount=amount,
            name="authorize",
            url=reverse("creditcards-authorize"),
            fields=fields,
        )

    def record_successful_authorization(
        self,
        order: "Order",
        amount: Decimal,
        reference: str,
    ) -> Complete:
        """Payment Step 3 Succeeded"""
        source = self.get_source(order, reference)

        source.allocate(amount, reference=reference, status="ACCEPTED")
        event = self.make_authorize_event(order, amount, reference)
        for line in order.lines.all():
            self.make_event_quantity(event, line, line.quantity)

        return Complete(amount, source_id=source.pk)

    def record_declined_authorization(
        self,
        order: "Order",
        amount: Decimal,
        reference: str,
    ) -> Declined:
        """Payment Step 3 Failed"""
        source = self.get_source(order, reference)
        source._create_transaction(Transaction.AUTHORISE, amount, reference=reference, status="DECLINED")
        return Declined(amount, source_id=source.pk)



# Generated by Django 3.1.6 on 2021-02-01 17:35

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("order", "0009_surcharge"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="orderdiscount",
            options={
                "ordering": ["pk"],
                "verbose_name": "Order Discount",
                "verbose_name_plural": "Order Discounts",
            },
        ),
        migrations.AlterModelOptions(
            name="ordernote",
            options={
                "ordering": ["-date_updated"],
                "verbose_name": "Order Note",
                "verbose_name_plural": "Order Notes",
            },
        ),
        migrations.RemoveField(
            model_name="line",
            name="est_dispatch_date",
        ),
        migrations.RemoveField(
            model_name="line",
            name="unit_cost_price",
        ),
        migrations.RemoveField(
            model_name="line",
            name="unit_retail_price",
        ),
    ]



from django.apps import apps
from django.conf import settings
from django.conf.urls import i18n
from django.contrib import admin
from django.urls import include, path
from django.views.static import serve

urlpatterns = [
    path("i18n/", include(i18n)),
    path("admin/", admin.site.urls),
    path(
        "media/<path:path>",
        serve,
        {"document_root": settings.MEDIA_ROOT, "show_indexes": True},
    ),
    # Include plugins
    path(
        "api/",
        include(
            apps.get_app_config("oscarapicheckout").urls[0]  # type:ignore[attr-defined]
        ),
    ),
    path("api/", include("oscarapi.urls")),
    path("creditcards/", include("sandbox.creditcards.urls")),
    path("clientside/", include("sandbox.clientside.urls")),
    # Include stock Oscar
    path(
        "",
        include(apps.get_app_config("oscar").urls[0]),  # type:ignore[attr-defined]
    ),
]



from typing import Any, TypedDict

from django.conf import settings


def overridable(name: str, default: Any | None = None) -> Any:
    return getattr(settings, name, default)


class PaymentMethodConfig(TypedDict):
    method: str
    permission: str
    method_kwargs: dict[str, Any]
    permission_kwargs: dict[str, Any]


class FraudRuleConfig(TypedDict):
    rule: str
    kwargs: dict[str, Any]


API_ENABLED_PAYMENT_METHODS: list[PaymentMethodConfig] = overridable(
    "API_ENABLED_PAYMENT_METHODS",
    [
        {
            "method": "oscarapicheckout.methods.Cash",
            "permission": "oscarapicheckout.permissions.StaffOnly",
            "method_kwargs": {},
            "permission_kwargs": {},
        }
    ],
)
API_MAX_PAYMENT_METHODS: int = overridable("API_MAX_PAYMENT_METHODS", 0)
API_CHECKOUT_CAPTCHA: str | bool = overridable(
    "API_CHECKOUT_CAPTCHA",
    "oscarapicheckout.utils.get_checkout_captcha_settings",
)
API_CHECKOUT_FRAUD_CHECKS: list[FraudRuleConfig] = overridable(
    "API_CHECKOUT_FRAUD_CHECKS",
    [],
)

ORDER_STATUS_PENDING: str = overridable("ORDER_STATUS_PENDING", "Pending")
ORDER_STATUS_PAYMENT_DECLINED: str = overridable("ORDER_STATUS_PAYMENT_DECLINED", "Payment Declined")
ORDER_STATUS_AUTHORIZED: str = overridable("ORDER_STATUS_AUTHORIZED", "Authorized")
ORDER_STATUS_SHIPPED: str = overridable("ORDER_STATUS_SHIPPED", "Shipped")
ORDER_STATUS_CANCELED: str = overridable("ORDER_STATUS_CANCELED", "Canceled")

ORDER_OWNERSHIP_CALCULATOR: str = overridable(
    "ORDER_OWNERSHIP_CALCULATOR",
    "oscarapicheckout.utils.get_order_ownership",
)

CHECKOUT_CACHE_SERIALIZERS: dict[str, str] = overridable(
    "CHECKOUT_CACHE_SERIALIZERS",
    {},
)



# Generated by Django 2.2.5 on 2019-09-26 11:52

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("order", "0006_auto_20190926_1152"),
        ("communication", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="communicationevent",
            name="event_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="communication.CommunicationEventType",
                verbose_name="Event Type",
            ),
        ),
    ]



from decimal import Decimal
from typing import Any, cast

from django.conf import settings
from django.contrib.auth.base_user import AbstractBaseUser
from django.contrib.auth.models import AnonymousUser
from django.db import transaction
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _
from oscar.apps.order.signals import order_placed
from oscar.core.loading import get_class, get_model
from oscar.core.prices import Price

Basket = get_model("basket", "Basket")
Order = get_model("order", "Order")
ShippingAddress = get_model("order", "ShippingAddress")
BillingAddress = get_model("order", "BillingAddress")

OrderCreator = get_class("order.utils", "OrderCreator")
OrderNumberGenerator = get_class("order.utils", "OrderNumberGenerator")
ShippingMethod = get_class("shipping.methods", "Base")


class OrderCreatorMixin(OrderCreator):
    """
    Class mixin for oscar.apps.order.utils.OrderCreator. Handles bug that occurs when
    order ownership isn't the same as the user placing the order, but the basket contains
    a voucher which is only available to the user placing the order.
    """

    def place_order(
        self,
        basket: Basket,
        total: Price,
        shipping_method: ShippingMethod,
        shipping_charge: Price,
        user: AbstractBaseUser | AnonymousUser | None = None,
        shipping_address: ShippingAddress | None = None,
        billing_address: BillingAddress | None = None,
        order_number: str | int | None = None,
        status: str | None = None,
        request: HttpRequest | None = None,
        surcharges: list[Any] | None = None,
        **kwargs: Any,
    ) -> Order:
        """
        Placing an order involves creating all the relevant models based on the
        basket and session data.
        """
        # Make sure basket isn't empty
        if basket.is_empty:
            # Translators: User facing error message in checkout
            raise ValueError(_("Empty baskets cannot be submitted"))

        # Allocate an order number
        if not order_number:
            generator = OrderNumberGenerator()
            order_number = generator.order_number(basket)

        # Figure out what status the new order should be
        if not status and hasattr(settings, "OSCAR_INITIAL_ORDER_STATUS"):
            status = getattr(settings, "OSCAR_INITIAL_ORDER_STATUS")

        # Make sure there isn't already an order with this order number
        if Order._default_manager.filter(number=order_number).exists():
            # Translators: User facing error message in checkout
            raise ValueError(_("There is already an order with number %(order_number)s") % dict(order_number=order_number))

        # Open a transaction so that order creation is atomic.
        with transaction.atomic():
            # Create the actual order.Order and order.Line models
            order = cast(
                Order,
                self.create_order_model(
                    user,
                    basket,
                    shipping_address,
                    shipping_method,
                    shipping_charge,
                    billing_address,
                    total,
                    order_number,
                    status,
                    request=request,
                    **kwargs,
                ),
            )
            for line in basket.all_lines():
                self.create_line_models(order, line)
                self.update_stock_records(line)

            # Make sure all the vouchers in the order are active and can actually be used by the order placing user.
            voucher_user = request.user if request and request.user else user
            for voucher in basket.vouchers.select_for_update():
                available_to_user, msg = voucher.is_available_to_user(user=voucher_user)
                if not voucher.is_active() or not available_to_user:
                    raise ValueError(msg)

            # Record any discounts associated with this order
            for application in basket.offer_applications:
                # Trigger any deferred benefits from offers and capture the resulting message
                application["message"] = application["offer"].apply_deferred_benefit(basket, order, application)
                # Record offer application results
                if application["result"].affects_shipping:
                    # Skip zero shipping discounts
                    shipping_discount = shipping_method.discount(basket)
                    if shipping_discount <= Decimal("0.00"):
                        continue
                    # If a shipping offer, we need to grab the actual discount off
                    # the shipping method instance, which should be wrapped in an
                    # OfferDiscount instance.
                    application["discount"] = shipping_discount
                self.create_discount_model(order, application)
                self.record_discount(application)

            # Record voucher usage for this order
            for voucher in basket.vouchers.all():
                self.record_voucher_usage(order, voucher, user)

        # Send signal for analytics to pick up
        order_placed.send(sender=self, order=order, user=user)

        # Done! Return the order.Order model
        return order



from datetime import timedelta
from decimal import Decimal

from django.utils import timezone
from oscar.core.loading import get_class, get_model
from rest_framework import serializers

from .. import fraud
from .base import BaseTest

Order = get_model("order", "Order")
ShippingAddress = get_model("order", "ShippingAddress")
Basket = get_model("basket", "Basket")
Default = get_class("partner.strategy", "Default")
OrderCreator = get_class("order.utils", "OrderCreator")
Country = get_model("address", "Country")


class AddressVelocityTest(BaseTest):
    def test_validate_address_velocity(self):
        # Create shipping address data
        address_data = {
            "title": "Mr",
            "first_name": "John",
            "last_name": "Doe",
            "line1": "123 Test St",
            "line2": "Test Building",
            "line3": "Test Floor",
            "line4": "Test City",
            "state": "Test State",
            "postcode": "TEST123",
            "country": Country.objects.get(pk="US"),
            "phone_number": "1234567890",
        }
        checkout_data = {
            "shipping_address": address_data,
            "billing_address": address_data,
        }

        # Address should pass check at first
        fraud.run_enabled_fraud_checks(checkout_data)

        # Create fake test orders with the same shipping address data
        for i in range(10):  # Create 10 orders
            # Create a fake test order with the specified shipping address
            Order.objects.create(
                number=str(1 + i),
                currency="USD",
                total_incl_tax=Decimal("100.00"),
                total_excl_tax=Decimal("90.00"),
                shipping_address=ShippingAddress.objects.create(**address_data),
                date_placed=timezone.now() - timedelta(hours=23),
            )

        # Address should now fail the fraud check
        with self.assertRaises(serializers.ValidationError):
            fraud.run_enabled_fraud_checks(checkout_data)



from django.contrib.auth.models import User
from oscar.core.loading import get_model
from rest_framework.test import APITestCase

Country = get_model("address", "Country")


class BaseTest(APITestCase):
    def setUp(self):
        Country.objects.create(
            display_order=0,
            is_shipping_country=True,
            iso_3166_1_a2="US",
            iso_3166_1_a3="USA",
            iso_3166_1_numeric="840",
            name="United States of America",
            printable_name="United States",
        )

    def login(self, is_staff=False, email="joe@example.com"):
        user = User.objects.create_user(username="joe", password="schmoe", email=email)
        user.is_staff = is_staff
        user.save()
        self.client.login(username="joe", password="schmoe")
        return user



from decimal import Decimal
from typing import Any, NotRequired, TypedDict
import logging

from django.db import transaction
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext_noop
from django_stubs_ext import StrOrPromise
from oscar.core.loading import get_model
from rest_framework import serializers

from . import states

Order = get_model("order", "Order")
OrderLine = get_model("order", "Line")
PaymentEventType = get_model("order", "PaymentEventType")
PaymentEvent = get_model("order", "PaymentEvent")
PaymentEventQuantity = get_model("order", "PaymentEventQuantity")
SourceType = get_model("payment", "SourceType")
Source = get_model("payment", "Source")
Transaction = get_model("payment", "Transaction")

logger = logging.getLogger(__name__)


class PaymentMethodData(TypedDict):
    method_type: str
    enabled: NotRequired[bool]
    pay_balance: NotRequired[bool]
    amount: NotRequired[Decimal]
    reference: NotRequired[str]


class PaymentMethodSerializer[T: PaymentMethodData](serializers.Serializer[Any]):
    method_type = serializers.ChoiceField(choices=tuple())
    enabled = serializers.BooleanField(default=False)
    pay_balance = serializers.BooleanField(default=True)
    amount = serializers.DecimalField(decimal_places=2, max_digits=12, required=False)
    reference = serializers.CharField(max_length=128, default="")

    def __init__(
        self,
        *args: Any,
        method_type_choices: list[tuple[str, str]] = list(),
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self.fields["method_type"] = serializers.ChoiceField(choices=method_type_choices)

    def validate(self, data: T) -> T:
        if not data["enabled"]:
            return data
        if data["pay_balance"]:
            data.pop("amount", None)  # type:ignore[arg-type]
        elif "amount" not in data or not data["amount"] or data["amount"] <= Decimal("0.00"):
            # Translators: User facing error message in checkout
            raise serializers.ValidationError(_("Amount must be greater then 0.00 or pay_balance must be enabled."))
        return data


class PaymentMethod[T: PaymentMethodData]:
    # Translators: Description of payment method in checkout
    name: StrOrPromise = gettext_noop("Abstract Payment Method")
    code: str = "abstract-payment-method"
    serializer_class: type[PaymentMethodSerializer[T]] = PaymentMethodSerializer

    def _make_payment_event(
        self,
        type_name: str,
        order: Order,
        amount: Decimal,
        reference: str = "",
    ) -> PaymentEvent:
        etype, created = PaymentEventType.objects.get_or_create(name=type_name)
        event = PaymentEvent()
        event.order = order
        event.amount = amount
        event.reference = reference if reference else ""
        event.event_type = etype
        event.save()
        return event

    def make_authorize_event(
        self,
        order: Order,
        amount: Decimal,
        reference: str = "",
    ) -> PaymentEvent:
        return self._make_payment_event(
            type_name=Transaction.AUTHORISE,
            order=order,
            amount=amount,
            reference=reference,
        )

    def make_debit_event(
        self,
        order: Order,
        amount: Decimal,
        reference: str = "",
    ) -> PaymentEvent:
        return self._make_payment_event(
            type_name=Transaction.DEBIT,
            order=order,
            amount=amount,
            reference=reference,
        )

    def make_refund_event(
        self,
        order: Order,
        amount: Decimal,
        reference: str = "",
    ) -> PaymentEvent:
        return self._make_payment_event(
            type_name=Transaction.REFUND,
            order=order,
            amount=amount,
            reference=reference,
        )

    def make_event_quantity(
        self,
        event: PaymentEvent,
        line: OrderLine,
        quantity: int,
    ) -> PaymentEventQuantity:
        return PaymentEventQuantity.objects.create(event=event, line=line, quantity=quantity)

    def get_source(
        self,
        order: Order,
        reference: str = "",
    ) -> Source:
        stype, created = SourceType.objects.get_or_create(name=self.name)
        source, created = Source.objects.get_or_create(order=order, source_type=stype, reference=reference)
        source.currency = order.currency
        source.save()
        return source

    @transaction.atomic()
    def void_existing_payment(
        self,
        request: HttpRequest,
        order: Order,
        method_key: str,
        state_to_void: states.PaymentStatus,
    ) -> None:
        source_id: int | None = getattr(state_to_void, "source_id", None)
        source = Source.objects.filter(pk=source_id).first() if source_id is not None else None
        if not source:
            logger.warning(
                "Attempted to void PaymentSource for Order[%s], MethodKey[%s], but no source was found.",
                order.number,
                method_key,
            )
            return
        source.amount_allocated = max(
            Decimal("0.00"),
            (source.amount_allocated - state_to_void.amount),
        )
        source.save()
        logger.info(
            "Voided Amount[%s] from PaymentSource[%s] for Order[%s], MethodKey[%s].",
            state_to_void.amount,
            source,
            order.number,
            method_key,
        )

    @transaction.atomic()
    def record_payment(
        self,
        request: HttpRequest,
        order: Order,
        method_key: str,
        amount: Decimal | None = None,
        reference: str = "",
        **kwargs: Any,
    ) -> states.PaymentStatus:
        if not amount and amount != Decimal("0.00"):
            raise RuntimeError("Amount must be specified")
        return self._record_payment(
            request,
            order,
            method_key,
            amount=amount,
            reference=reference,
            **kwargs,
        )

    def _record_payment(
        self,
        request: HttpRequest,
        order: Order,
        method_key: str,
        amount: Decimal,
        reference: str,
        **kwargs: Any,
    ) -> states.PaymentStatus:
        raise NotImplementedError("Subclass must implement _record_payment(request, order, method_key, amount, reference, **kwargs) method.")


class Cash(PaymentMethod[PaymentMethodData]):
    """
    Cash payments are an example of how to implement a payment method plug-in. It
    doesn't do anything more than record a transaction and payment source.
    """

    # Translators: Description of payment method in checkout
    name = gettext_noop("Cash")
    code = "cash"

    def _record_payment(
        self,
        request: HttpRequest,
        order: Order,
        method_key: str,
        amount: Decimal,
        reference: str,
        **kwargs: Any,
    ) -> states.PaymentStatus:
        source = self.get_source(order, reference)

        amount_to_allocate = amount - source.amount_allocated
        source.allocate(amount_to_allocate, reference)

        amount_to_debit = amount - source.amount_debited
        source.debit(amount_to_debit, reference)

        event = self.make_debit_event(order, amount_to_debit, reference)
        for line in order.lines.all():
            self.make_event_quantity(event, line, line.quantity)

        return states.Complete(source.amount_debited, source_id=source.pk)


class PayLater(PaymentMethod[PaymentMethodData]):
    """
    The PayLater method method is similar to the Cash payment method—it doesn't
    actually authorize any payment. This method exists to differentiate two
    different types of not-really-authorized orders:

    - `Cash` is intended for use by employees when they are going to collect
        payment directly by some means other than Oscar.
    - `PayLater` is used by customers as an failover system when the normal
        payment processor is experiencing downtime. It allows a customer to
        place an order, then come back and complete the payment authorization
        later.
    """

    # Translators: Description of payment method in checkout
    name = gettext_noop("Pay Later")
    code = "pay-later"

    def _record_payment(
        self,
        request: HttpRequest,
        order: Order,
        method_key: str,
        amount: Decimal,
        reference: str,
        **kwargs: Any,
    ) -> states.PaymentStatus:
        source = self.get_source(order, reference)
        return states.Deferred(Decimal("0.00"), source_id=source.pk)



from django.urls import path

from .views import CompleteClientSidePaymentView

urlpatterns = [
    path(
        "complete/",
        CompleteClientSidePaymentView.as_view(),
        name="clientside-complete",
    ),
]



from decimal import Decimal
from typing import Any, TypedDict
import base64
import pickle

from django.contrib.auth.base_user import AbstractBaseUser
from django.contrib.auth.models import AnonymousUser, User
from django.db.models import F
from django.db.models.functions import Greatest
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _
from oscar.core.loading import get_class, get_model
from oscar.core.prices import Price
from oscarapi.basket import operations

from .settings import ORDER_STATUS_AUTHORIZED, ORDER_STATUS_PAYMENT_DECLINED
from .signals import order_payment_authorized, order_payment_declined
from .states import Complete, Consumed, Declined, PaymentMethodStatus, PaymentStatus

Basket = get_model("basket", "Basket")
Order = get_model("order", "Order")
ShippingAddress = get_model("order", "ShippingAddress")
BillingAddress = get_model("order", "BillingAddress")

OrderCreator = get_class("order.utils", "OrderCreator")
ShippingMethod = get_class("shipping.methods", "Base")

CHECKOUT_ORDER_ID = "api_checkout_pending_order_id"
CHECKOUT_PAYMENT_STEPS = "api_checkout_payment_steps"


def _session_pickle(obj: Any) -> str:
    pickled = pickle.dumps(obj)
    base64ed = base64.standard_b64encode(pickled)
    utfed = base64ed.decode("utf8")
    return utfed


def _session_unpickle(utfed: str) -> Any:
    base64ed = utfed.encode("utf8")
    pickled = base64.standard_b64decode(base64ed)
    obj = pickle.loads(pickled)
    return obj


def _update_payment_method_state(
    request: HttpRequest,
    method_key: str,
    state: PaymentStatus,
) -> None:
    states = request.session.get(CHECKOUT_PAYMENT_STEPS, {})
    states[method_key] = _session_pickle(state)
    request.session[CHECKOUT_PAYMENT_STEPS] = states
    request.session.modified = True


def _set_order_authorized(order: Order, request: HttpRequest) -> None:
    # Set the order status
    order.set_status(ORDER_STATUS_AUTHORIZED)

    if order.basket is not None:
        # Mark the basket as submitted
        order.basket.submit()

        # Update the owner of the basket to match the order
        if order.user != order.basket.owner:
            order.basket.owner = order.user
            order.basket.save()

    # Send a signal
    order_payment_authorized.send(sender=order, order=order, request=request)


def _set_order_payment_declined(order: Order, request: HttpRequest) -> None:
    # Set the order status
    order.set_status(ORDER_STATUS_PAYMENT_DECLINED)

    voucher_applications = order.voucherapplication_set.all()

    for voucher_application in voucher_applications:
        voucher = voucher_application.voucher

        parent = getattr(voucher, "parent", None)
        if parent:
            parent.num_orders = Greatest(F("num_orders") - 1, 0)
            parent.save(update_children=False)

        voucher.num_orders = Greatest(F("num_orders") - 1, 0)
        voucher.save()

    # Delete some related objects
    order.discounts.all().delete()
    order.line_prices.all().delete()
    voucher_applications.delete()

    if order.basket is not None:
        # Thaw the basket and put it back into the request.session so that it can be retried
        order.basket.thaw()
        operations.store_basket_in_session(order.basket, request.session)

    # Send a signal
    order_payment_declined.send(sender=order, order=order, request=request)


def _update_order_status(order: Order, request: HttpRequest) -> None:
    states = list_payment_method_states(request)

    declined = [s for k, s in states.items() if s.status == PaymentMethodStatus.DECLINED]
    if len(declined) > 0:
        _set_order_payment_declined(order, request)

    not_complete = [s for k, s in states.items() if s.status != PaymentMethodStatus.COMPLETE]
    if len(not_complete) <= 0:
        # Authorized the order and consume all the payments
        _set_order_authorized(order, request)
        for key, state in states.items():
            mark_payment_method_consumed(
                order,
                request,
                key,
                state.amount,
                source_id=getattr(state, "source_id", None),
            )


def list_payment_method_states(request: HttpRequest) -> dict[str, PaymentStatus]:
    states = request.session.get(CHECKOUT_PAYMENT_STEPS, {})
    return {method_key: _session_unpickle(state) for method_key, state in states.items()}


def clear_payment_method_states(request: HttpRequest) -> None:
    request.session[CHECKOUT_PAYMENT_STEPS] = {}
    request.session.modified = True


def clear_consumed_payment_method_states(request: HttpRequest) -> None:
    curr_states = list_payment_method_states(request)
    new_states = {}
    for key, state in curr_states.items():
        if state.status != PaymentMethodStatus.CONSUMED:
            new_states[key] = state
    clear_payment_method_states(request)
    for key, state in new_states.items():
        _update_payment_method_state(request, key, state)


def update_payment_method_state(
    order: Order,
    request: HttpRequest,
    method_key: str,
    state: PaymentStatus,
) -> None:
    _update_payment_method_state(request, method_key, state)
    _update_order_status(order, request)


def set_payment_method_states(
    order: Order,
    request: HttpRequest,
    states: dict[str, PaymentStatus],
) -> None:
    clear_payment_method_states(request)
    for method_key, state in states.items():
        _update_payment_method_state(request, method_key, state)
    _update_order_status(order, request)


def mark_payment_method_completed(
    order: Order,
    request: HttpRequest,
    method_key: str,
    amount: Decimal,
    source_id: int | None = None,
) -> None:
    update_payment_method_state(
        order,
        request,
        method_key,
        Complete(amount, source_id=source_id),
    )


def mark_payment_method_declined(
    order: Order,
    request: HttpRequest,
    method_key: str,
    amount: Decimal,
    source_id: int | None = None,
) -> None:
    update_payment_method_state(
        order,
        request,
        method_key,
        Declined(amount, source_id=source_id),
    )


def mark_payment_method_consumed(
    order: Order,
    request: HttpRequest,
    method_key: str,
    amount: Decimal,
    source_id: int | None = None,
) -> None:
    update_payment_method_state(
        order,
        request,
        method_key,
        Consumed(amount, source_id=source_id),
    )


def get_order_ownership(
    request: HttpRequest,
    given_user: User | None,
    guest_email: str | None,
) -> tuple[User | None, str | None]:
    current_user = request.user
    if current_user and current_user.is_authenticated:
        return current_user, None
    return None, guest_email


class CheckoutCaptchaSettings(TypedDict):
    action: str
    required_score: float


def get_checkout_captcha_settings(
    request: HttpRequest,
) -> None | CheckoutCaptchaSettings:
    return None


class OrderUpdater:
    def update_order(
        self,
        order: Order,
        basket: Basket,
        order_total: Price,
        shipping_method: ShippingMethod,
        shipping_charge: Price,
        user: User | AnonymousUser | None = None,
        shipping_address: ShippingAddress | None = None,
        billing_address: BillingAddress | None = None,
        order_number: str | None = None,
        status: str | None = None,
        request: HttpRequest | None = None,
        **kwargs: Any,
    ) -> Order:
        """
        Similar to OrderCreator.place_order, except this updates an existing "Payment Declined" order instead
        of creating a new order.
        """
        if basket.is_empty:
            # Translators: Error message in checkout
            raise ValueError(_("Empty baskets cannot be submitted"))

        if order.status != ORDER_STATUS_PAYMENT_DECLINED:
            # Translators: Error message in checkout
            raise ValueError(_("Can not update an order that isn't in payment declined state."))

        # Make sure there isn't another order with this number already, besides of course the
        # order we're trying to update.
        try:
            Order._default_manager.exclude(id=order.id).get(number=order_number)
        except Order.DoesNotExist:
            pass
        else:
            # Translators: Error message in checkout
            msg = _("There is already an order with number %(order_number)s") % dict(order_number=order_number)
            raise ValueError(msg)

        # Remove all the order lines and cancel and stock they allocated. We'll make new lines from the
        # basket after this.
        for order_line in order.lines.all():
            product_class = order_line.product.get_product_class() if order_line.product else None
            if product_class and product_class.track_stock and order_line.stockrecord:
                order_line.stockrecord.cancel_allocation(order_line.quantity)
            order_line.delete()

        # Use the built in OrderCreator, but specify a pk so that Django actually does an update instead
        # Create the actual order.Order and order.Line models
        if order_number is None:
            raise ValueError("order_number is required")
        order_user: AbstractBaseUser | None = user if isinstance(user, AbstractBaseUser) else None
        creator = OrderCreator()
        order = creator.create_order_model(
            order_user,
            basket,
            shipping_address,
            shipping_method,
            shipping_charge,
            billing_address,
            order_total,
            order_number,
            status,
            id=order.id,
            request=request,
            **kwargs,
        )

        # Make new order lines to replace the ones we deleted.
        for basket_line in basket.all_lines():
            creator.create_line_models(order, basket_line)
            creator.update_stock_records(basket_line)

        # Make sure all the vouchers are still available to the user placing the order (not necessarily the
        # same as the order owner)
        voucher_user = request.user if request and request.user else user
        for voucher in basket.vouchers.select_for_update():
            available_to_user, msg = voucher.is_available_to_user(user=voucher_user)
            if not voucher.is_active() or not available_to_user:
                raise ValueError(msg)

        # Record any discounts associated with this order
        for application in basket.offer_applications:
            # Trigger any deferred benefits from offers and capture the
            # resulting message
            application["message"] = application["offer"].apply_deferred_benefit(basket, order, application)

            # Record offer application results
            if application["result"].affects_shipping:
                # Skip zero shipping discounts
                shipping_discount = shipping_method.discount(basket)
                if shipping_discount <= Decimal("0.00"):
                    continue
                # If a shipping offer, we need to grab the actual discount off
                # the shipping method instance, which should be wrapped in an
                # OfferDiscount instance.
                application["discount"] = shipping_discount
            creator.create_discount_model(order, application)
            creator.record_discount(application)

        # Record voucher usage for this order
        for voucher in basket.vouchers.all():
            creator.record_voucher_usage(order, voucher, user)

        # Done! Return the order.Order model
        return order



from decimal import Decimal
from typing import TYPE_CHECKING, Any

from django.core.signing import Signer
from django.shortcuts import get_object_or_404
from oscar.core.loading import get_model
from oscarapicheckout.states import PaymentStatus
from rest_framework import generics
from rest_framework.request import Request
from rest_framework.response import Response

from oscarapicheckout import utils

from .methods import ClientSideCard

if TYPE_CHECKING:
    from ..order.models import Order
else:
    Order = get_model("order", "Order")


class CompleteClientSidePaymentView(generics.GenericAPIView[Any]):
    def post(self, request: Request) -> Response:
        amount = Decimal(request.data["amount"])
        order_number = request.data["reference_number"]
        order = get_object_or_404(Order, number=order_number)

        method_key = Signer().unsign(request.data["transaction_id"])

        new_state: PaymentStatus
        if request.data.get("deny"):
            new_state = ClientSideCard().record_declined_authorization(order, amount, reference="")
            utils.update_payment_method_state(order, request, method_key, new_state)
            return Response({"status": "Declined"})

        reference = request.data.get("result_token", "")
        new_state = ClientSideCard().record_successful_authorization(order, amount, reference)
        utils.update_payment_method_state(order, request, method_key, new_state)
        return Response({"status": "Success"})



from django.urls import path
from django.urls.resolvers import URLPattern, URLResolver
from django.utils.translation import gettext_lazy as _
from django.views.decorators.cache import never_cache
from oscar.core.application import OscarConfig


class Config(OscarConfig):
    name = "oscarapicheckout"
    label = "oscarapicheckout"
    verbose_name = _("Oscar API-Checkout")
    namespace = "oscarapicheckout"
    default = True

    def ready(self) -> None:
        # Register signal handlers
        from . import handlers  # NOQA

    def get_urls(self) -> list[URLPattern | URLResolver]:
        from .views import (
            CheckoutView,
            CompleteDeferredPaymentView,
            PaymentMethodsView,
            PaymentStatesView,
        )

        view_methods = never_cache(PaymentMethodsView.as_view())
        view_states = never_cache(PaymentStatesView.as_view())
        view_checkout = never_cache(CheckoutView.as_view())
        view_complete_deferred_payment = never_cache(CompleteDeferredPaymentView.as_view())
        urlpatterns: list[URLPattern | URLResolver] = [
            path(
                "checkout/payment-methods/",
                view_methods,
                name="api-checkout-payment-methods",
            ),
            path("checkout/payment-states/", view_states, name="api-payment"),
            path(
                "checkout/payment-states/<int:pk>/",
                view_states,
                name="api-payment",
            ),
            path(
                "checkout/complete-deferred-payment/",
                view_complete_deferred_payment,
                name="api-complete-deferred-payment",
            ),
            path("checkout/", view_checkout, name="api-checkout"),
        ]
        return self.post_process_urls(urlpatterns)



from django.urls import path

from .views import AuthorizeCardView, GetCardTokenView

urlpatterns = [
    path("get-token/", GetCardTokenView.as_view(), name="creditcards-get-token"),
    path("authorize/", AuthorizeCardView.as_view(), name="creditcards-authorize"),
]



from decimal import Decimal

from oscar.test.factories import create_order

from ..methods import Cash, PaymentMethod, PaymentMethodSerializer
from .base import BaseTest


class PaymentMethodSerializerTest(BaseTest):
    def test_disabled_is_valid(self):
        s = PaymentMethodSerializer(
            method_type_choices=[("cash", "Cash")],
            data={
                "method_type": "cash",
                "enabled": False,
            },
        )
        self.assertTrue(s.is_valid())
        self.assertDictEqual(
            s.validated_data,
            {
                "method_type": "cash",
                "enabled": False,
                "pay_balance": True,
                "reference": "",
            },
        )

    def test_pay_balance_without_amount_is_valid(self):
        s = PaymentMethodSerializer(
            method_type_choices=[("cash", "Cash")],
            data={
                "method_type": "cash",
                "enabled": True,
                "pay_balance": True,
                "amount": "10.00",
            },
        )
        self.assertTrue(s.is_valid())
        self.assertDictEqual(
            s.validated_data,
            {
                "method_type": "cash",
                "enabled": True,
                "pay_balance": True,
                "reference": "",
            },
        )

    def test_amount_specified_is_valid(self):
        # amount specified without amount is valid
        s = PaymentMethodSerializer(
            method_type_choices=[("cash", "Cash")],
            data={
                "method_type": "cash",
                "enabled": True,
                "pay_balance": False,
                "amount": "10.00",
            },
        )
        self.assertTrue(s.is_valid())
        self.assertDictEqual(
            s.validated_data,
            {
                "method_type": "cash",
                "enabled": True,
                "pay_balance": False,
                "amount": Decimal("10.00"),
                "reference": "",
            },
        )

    def test_missing_amount_is_not_valid(self):
        s = PaymentMethodSerializer(
            method_type_choices=[("cash", "Cash")],
            data={
                "method_type": "cash",
                "enabled": True,
                "pay_balance": False,
            },
        )
        self.assertFalse(s.is_valid())

    def test_missing_method_type_is_not_valid(self):
        s = PaymentMethodSerializer(
            method_type_choices=[("cash", "Cash")],
            data={
                "enabled": True,
                "pay_balance": False,
            },
        )
        self.assertFalse(s.is_valid())

    def test_invalid_method_type_is_not_valid(self):
        s = PaymentMethodSerializer(
            method_type_choices=[("cash", "Cash")],
            data={
                "method_type": "credit-card",
                "enabled": True,
                "pay_balance": False,
            },
        )
        self.assertFalse(s.is_valid())

        s = PaymentMethodSerializer(
            method_type_choices=[],
            data={
                "method_type": "cash",
                "enabled": True,
                "pay_balance": False,
            },
        )
        self.assertFalse(s.is_valid())

        s = PaymentMethodSerializer(
            method_type_choices=["credit-card", "Credit Card"],
            data={
                "method_type": "cash",
                "enabled": True,
                "pay_balance": False,
            },
        )
        self.assertFalse(s.is_valid())


class PaymentMethodTest(BaseTest):
    def test_get_source(self):
        order = create_order()

        method = PaymentMethod()
        source = method.get_source(order, reference="12345")

        self.assertEqual(source.source_type.name, "Abstract Payment Method")
        self.assertEqual(source.order, order)
        self.assertEqual(source.currency, order.currency)
        self.assertEqual(source.reference, "12345")

        self.assertEqual(order.sources.all().count(), 1)
        self.assertEqual(order.sources.first().id, source.id)


class CashTest(BaseTest):
    def test_record_payment(self):
        order = create_order()
        state = Cash().record_payment(None, order, "cash", amount=Decimal("1.00"))

        self.assertEqual(state.status, "Complete")
        self.assertEqual(state.amount, Decimal("1.00"))

        self.assertEqual(order.sources.count(), 1)

        source = order.sources.first()
        self.assertEqual(source.currency, order.currency)
        self.assertEqual(source.amount_allocated, Decimal("1.00"))
        self.assertEqual(source.amount_debited, Decimal("1.00"))
        self.assertEqual(source.amount_refunded, Decimal("0.00"))
        self.assertEqual(source.transactions.count(), 2)

        transaction = source.transactions.get(txn_type="Authorise")
        self.assertEqual(transaction.amount, Decimal("1.00"))

        transaction = source.transactions.get(txn_type="Debit")
        self.assertEqual(transaction.amount, Decimal("1.00"))



from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import oscar.models.fields
import oscar.models.fields.autoslugfield


class Migration(migrations.Migration):
    dependencies = [
        ("partner", "0001_initial"),
        ("customer", "0001_initial"),
        ("address", "0001_initial"),
        ("basket", "0002_auto_20140827_1705"),
        ("catalogue", "0001_initial"),
        ("sites", "0001_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="BillingAddress",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "title",
                    models.CharField(
                        verbose_name="Title",
                        max_length=64,
                        blank=True,
                        choices=[
                            ("Mr", "Mr"),
                            ("Miss", "Miss"),
                            ("Mrs", "Mrs"),
                            ("Ms", "Ms"),
                            ("Dr", "Dr"),
                        ],
                    ),
                ),  # NOQA
                (
                    "first_name",
                    models.CharField(
                        max_length=255, verbose_name="First name", blank=True
                    ),
                ),
                (
                    "last_name",
                    models.CharField(
                        max_length=255, verbose_name="Last name", blank=True
                    ),
                ),
                (
                    "line1",
                    models.CharField(
                        max_length=255, verbose_name="First line of address"
                    ),
                ),
                (
                    "line2",
                    models.CharField(
                        max_length=255,
                        verbose_name="Second line of address",
                        blank=True,
                    ),
                ),
                (
                    "line3",
                    models.CharField(
                        max_length=255, verbose_name="Third line of address", blank=True
                    ),
                ),
                (
                    "line4",
                    models.CharField(max_length=255, verbose_name="City", blank=True),
                ),
                (
                    "state",
                    models.CharField(
                        max_length=255, verbose_name="State/County", blank=True
                    ),
                ),
                (
                    "postcode",
                    oscar.models.fields.UppercaseCharField(
                        max_length=64, verbose_name="Post/Zip-code", blank=True
                    ),
                ),
                (
                    "search_text",
                    models.TextField(
                        editable=False,
                        verbose_name="Search text - used only for searching addresses",
                    ),
                ),
                (
                    "country",
                    models.ForeignKey(
                        verbose_name="Country",
                        to="address.Country",
                        on_delete=models.CASCADE,
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "Billing addresses",
                "verbose_name": "Billing address",
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="CommunicationEvent",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "date_created",
                    models.DateTimeField(auto_now_add=True, verbose_name="Date"),
                ),
                (
                    "event_type",
                    models.ForeignKey(
                        verbose_name="Event Type",
                        to="customer.CommunicationEventType",
                        on_delete=models.CASCADE,
                    ),
                ),
            ],
            options={
                "ordering": ["-date_created"],
                "verbose_name_plural": "Communication Events",
                "verbose_name": "Communication Event",
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="Line",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "partner_name",
                    models.CharField(
                        max_length=128, verbose_name="Partner name", blank=True
                    ),
                ),
                (
                    "partner_sku",
                    models.CharField(max_length=128, verbose_name="Partner SKU"),
                ),
                (
                    "partner_line_reference",
                    models.CharField(
                        verbose_name="Partner reference",
                        max_length=128,
                        help_text="This is the item number that the partner uses within their system",
                        blank=True,
                    ),
                ),  # NOQA
                (
                    "partner_line_notes",
                    models.TextField(verbose_name="Partner Notes", blank=True),
                ),
                ("title", models.CharField(max_length=255, verbose_name="Title")),
                (
                    "upc",
                    models.CharField(
                        verbose_name="UPC", max_length=128, blank=True, null=True
                    ),
                ),
                (
                    "quantity",
                    models.PositiveIntegerField(default=1, verbose_name="Quantity"),
                ),
                (
                    "line_price_incl_tax",
                    models.DecimalField(
                        max_digits=12, decimal_places=2, verbose_name="Price (inc. tax)"
                    ),
                ),
                (
                    "line_price_excl_tax",
                    models.DecimalField(
                        max_digits=12,
                        decimal_places=2,
                        verbose_name="Price (excl. tax)",
                    ),
                ),
                (
                    "line_price_before_discounts_incl_tax",
                    models.DecimalField(
                        max_digits=12,
                        decimal_places=2,
                        verbose_name="Price before discounts (inc. tax)",
                    ),
                ),  # NOQA
                (
                    "line_price_before_discounts_excl_tax",
                    models.DecimalField(
                        max_digits=12,
                        decimal_places=2,
                        verbose_name="Price before discounts (excl. tax)",
                    ),
                ),  # NOQA
                (
                    "unit_cost_price",
                    models.DecimalField(
                        max_digits=12,
                        decimal_places=2,
                        blank=True,
                        verbose_name="Unit Cost Price",
                        null=True,
                    ),
                ),
                (
                    "unit_price_incl_tax",
                    models.DecimalField(
                        max_digits=12,
                        decimal_places=2,
                        blank=True,
                        verbose_name="Unit Price (inc. tax)",
                        null=True,
                    ),
                ),
                (
                    "unit_price_excl_tax",
                    models.DecimalField(
                        max_digits=12,
                        decimal_places=2,
                        blank=True,
                        verbose_name="Unit Price (excl. tax)",
                        null=True,
                    ),
                ),
                (
                    "unit_retail_price",
                    models.DecimalField(
                        max_digits=12,
                        decimal_places=2,
                        blank=True,
                        verbose_name="Unit Retail Price",
                        null=True,
                    ),
                ),
                (
                    "status",
                    models.CharField(max_length=255, verbose_name="Status", blank=True),
                ),
                (
                    "est_dispatch_date",
                    models.DateField(
                        blank=True, verbose_name="Estimated Dispatch Date", null=True
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "Order Lines",
                "verbose_name": "Order Line",
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="LineAttribute",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("type", models.CharField(max_length=128, verbose_name="Type")),
                ("value", models.CharField(max_length=255, verbose_name="Value")),
                (
                    "line",
                    models.ForeignKey(
                        verbose_name="Line",
                        related_name="attributes",
                        to="order.Line",
                        on_delete=models.CASCADE,
                    ),
                ),
                (
                    "option",
                    models.ForeignKey(
                        verbose_name="Option",
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="line_attributes",
                        to="catalogue.Option",
                        null=True,
                    ),
                ),  # NOQA
            ],
            options={
                "verbose_name_plural": "Line Attributes",
                "verbose_name": "Line Attribute",
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="LinePrice",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "quantity",
                    models.PositiveIntegerField(default=1, verbose_name="Quantity"),
                ),
                (
                    "price_incl_tax",
                    models.DecimalField(
                        max_digits=12, decimal_places=2, verbose_name="Price (inc. tax)"
                    ),
                ),
                (
                    "price_excl_tax",
                    models.DecimalField(
                        max_digits=12,
                        decimal_places=2,
                        verbose_name="Price (excl. tax)",
                    ),
                ),
                (
                    "shipping_incl_tax",
                    models.DecimalField(
                        default=0,
                        max_digits=12,
                        decimal_places=2,
                        verbose_name="Shiping (inc. tax)",
                    ),
                ),
                (
                    "shipping_excl_tax",
                    models.DecimalField(
                        default=0,
                        max_digits=12,
                        decimal_places=2,
                        verbose_name="Shipping (excl. tax)",
                    ),
                ),
                (
                    "line",
                    models.ForeignKey(
                        verbose_name="Line",
                        related_name="prices",
                        to="order.Line",
                        on_delete=models.CASCADE,
                    ),
                ),
            ],
            options={
                "ordering": ("id",),
                "verbose_name_plural": "Line Prices",
                "verbose_name": "Line Price",
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="Order",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "number",
                    models.CharField(
                        max_length=128,
                        unique=True,
                        db_index=True,
                        verbose_name="Order number",
                    ),
                ),
                (
                    "currency",
                    models.CharField(
                        default="GBP", max_length=12, verbose_name="Currency"
                    ),
                ),
                (
                    "total_incl_tax",
                    models.DecimalField(
                        max_digits=12,
                        decimal_places=2,
                        verbose_name="Order total (inc. tax)",
                    ),
                ),
                (
                    "total_excl_tax",
                    models.DecimalField(
                        max_digits=12,
                        decimal_places=2,
                        verbose_name="Order total (excl. tax)",
                    ),
                ),
                (
                    "shipping_incl_tax",
                    models.DecimalField(
                        default=0,
                        max_digits=12,
                        decimal_places=2,
                        verbose_name="Shipping charge (inc. tax)",
                    ),
                ),
                (
                    "shipping_excl_tax",
                    models.DecimalField(
                        default=0,
                        max_digits=12,
                        decimal_places=2,
                        verbose_name="Shipping charge (excl. tax)",
                    ),
                ),
                (
                    "shipping_method",
                    models.CharField(
                        max_length=128, verbose_name="Shipping method", blank=True
                    ),
                ),
                (
                    "shipping_code",
                    models.CharField(default="", max_length=128, blank=True),
                ),
                (
                    "status",
                    models.CharField(max_length=100, verbose_name="Status", blank=True),
                ),
                (
                    "guest_email",
                    models.EmailField(
                        max_length=75, verbose_name="Guest email address", blank=True
                    ),
                ),
                ("date_placed", models.DateTimeField(auto_now_add=True, db_index=True)),
                (
                    "basket",
                    models.ForeignKey(
                        null=True,
                        verbose_name="Basket",
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="basket.Basket",
                        blank=True,
                    ),
                ),
                (
                    "billing_address",
                    models.ForeignKey(
                        null=True,
                        verbose_name="Billing Address",
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="order.BillingAddress",
                        blank=True,
                    ),
                ),  # NOQA
            ],
            options={
                "ordering": ["-date_placed"],
                "verbose_name_plural": "Orders",
                "verbose_name": "Order",
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="OrderDiscount",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "category",
                    models.CharField(
                        default="Basket",
                        max_length=64,
                        verbose_name="Discount category",
                        choices=[
                            ("Basket", "Basket"),
                            ("Shipping", "Shipping"),
                            ("Deferred", "Deferred"),
                        ],
                    ),
                ),  # NOQA
                (
                    "offer_id",
                    models.PositiveIntegerField(
                        blank=True, verbose_name="Offer ID", null=True
                    ),
                ),
                (
                    "offer_name",
                    models.CharField(
                        max_length=128,
                        db_index=True,
                        verbose_name="Offer name",
                        blank=True,
                    ),
                ),
                (
                    "voucher_id",
                    models.PositiveIntegerField(
                        blank=True, verbose_name="Voucher ID", null=True
                    ),
                ),
                (
                    "voucher_code",
                    models.CharField(
                        max_length=128, db_index=True, verbose_name="Code", blank=True
                    ),
                ),
                (
                    "frequency",
                    models.PositiveIntegerField(verbose_name="Frequency", null=True),
                ),
                (
                    "amount",
                    models.DecimalField(
                        default=0,
                        max_digits=12,
                        decimal_places=2,
                        verbose_name="Amount",
                    ),
                ),
                ("message", models.TextField(blank=True)),
                (
                    "order",
                    models.ForeignKey(
                        verbose_name="Order",
                        related_name="discounts",
                        to="order.Order",
                        on_delete=models.CASCADE,
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "Order Discounts",
                "verbose_name": "Order Discount",
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="OrderNote",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "note_type",
                    models.CharField(
                        max_length=128, verbose_name="Note Type", blank=True
                    ),
                ),
                ("message", models.TextField(verbose_name="Message")),
                (
                    "date_created",
                    models.DateTimeField(
                        auto_now_add=True, verbose_name="Date Created"
                    ),
                ),
                (
                    "date_updated",
                    models.DateTimeField(auto_now=True, verbose_name="Date Updated"),
                ),
                (
                    "order",
                    models.ForeignKey(
                        verbose_name="Order",
                        related_name="notes",
                        to="order.Order",
                        on_delete=models.CASCADE,
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        verbose_name="User",
                        to=settings.AUTH_USER_MODEL,
                        null=True,
                        on_delete=models.CASCADE,
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "Order Notes",
                "verbose_name": "Order Note",
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="PaymentEvent",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "amount",
                    models.DecimalField(
                        max_digits=12, decimal_places=2, verbose_name="Amount"
                    ),
                ),
                (
                    "reference",
                    models.CharField(
                        max_length=128, verbose_name="Reference", blank=True
                    ),
                ),
                (
                    "date_created",
                    models.DateTimeField(
                        auto_now_add=True, verbose_name="Date created"
                    ),
                ),
            ],
            options={
                "ordering": ["-date_created"],
                "verbose_name_plural": "Payment Events",
                "verbose_name": "Payment Event",
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="PaymentEventQuantity",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("quantity", models.PositiveIntegerField(verbose_name="Quantity")),
                (
                    "event",
                    models.ForeignKey(
                        verbose_name="Event",
                        related_name="line_quantities",
                        to="order.PaymentEvent",
                        on_delete=models.CASCADE,
                    ),
                ),
                (
                    "line",
                    models.ForeignKey(
                        verbose_name="Line",
                        related_name="payment_event_quantities",
                        to="order.Line",
                        on_delete=models.CASCADE,
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "Payment Event Quantities",
                "verbose_name": "Payment Event Quantity",
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="PaymentEventType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(unique=True, max_length=128, verbose_name="Name"),
                ),
                (
                    "code",
                    oscar.models.fields.autoslugfield.AutoSlugField(
                        populate_from="name",
                        unique=True,
                        verbose_name="Code",
                        max_length=128,
                        editable=False,
                        blank=True,
                    ),
                ),  # NOQA
            ],
            options={
                "ordering": ("name",),
                "verbose_name_plural": "Payment Event Types",
                "verbose_name": "Payment Event Type",
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="ShippingAddress",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "title",
                    models.CharField(
                        verbose_name="Title",
                        max_length=64,
                        blank=True,
                        choices=[
                            ("Mr", "Mr"),
                            ("Miss", "Miss"),
                            ("Mrs", "Mrs"),
                            ("Ms", "Ms"),
                            ("Dr", "Dr"),
                        ],
                    ),
                ),  # NOQA
                (
                    "first_name",
                    models.CharField(
                        max_length=255, verbose_name="First name", blank=True
                    ),
                ),
                (
                    "last_name",
                    models.CharField(
                        max_length=255, verbose_name="Last name", blank=True
                    ),
                ),
                (
                    "line1",
                    models.CharField(
                        max_length=255, verbose_name="First line of address"
                    ),
                ),
                (
                    "line2",
                    models.CharField(
                        max_length=255,
                        verbose_name="Second line of address",
                        blank=True,
                    ),
                ),
                (
                    "line3",
                    models.CharField(
                        max_length=255, verbose_name="Third line of address", blank=True
                    ),
                ),
                (
                    "line4",
                    models.CharField(max_length=255, verbose_name="City", blank=True),
                ),
                (
                    "state",
                    models.CharField(
                        max_length=255, verbose_name="State/County", blank=True
                    ),
                ),
                (
                    "postcode",
                    oscar.models.fields.UppercaseCharField(
                        max_length=64, verbose_name="Post/Zip-code", blank=True
                    ),
                ),
                (
                    "search_text",
                    models.TextField(
                        editable=False,
                        verbose_name="Search text - used only for searching addresses",
                    ),
                ),
                (
                    "phone_number",
                    oscar.models.fields.PhoneNumberField(
                        verbose_name="Phone number",
                        help_text="In case we need to call you about your order",
                        blank=True,
                    ),
                ),  # NOQA
                (
                    "notes",
                    models.TextField(
                        verbose_name="Instructions",
                        help_text="Tell us anything we should know when delivering your order.",
                        blank=True,
                    ),
                ),
                (
                    "country",
                    models.ForeignKey(
                        verbose_name="Country",
                        to="address.Country",
                        on_delete=models.CASCADE,
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "Shipping addresses",
                "verbose_name": "Shipping address",
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="ShippingEvent",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "notes",
                    models.TextField(
                        verbose_name="Event notes",
                        help_text="This could be the dispatch reference, or a tracking number",
                        blank=True,
                    ),
                ),
                (
                    "date_created",
                    models.DateTimeField(
                        auto_now_add=True, verbose_name="Date Created"
                    ),
                ),
            ],
            options={
                "ordering": ["-date_created"],
                "verbose_name_plural": "Shipping Events",
                "verbose_name": "Shipping Event",
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="ShippingEventQuantity",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("quantity", models.PositiveIntegerField(verbose_name="Quantity")),
                (
                    "event",
                    models.ForeignKey(
                        verbose_name="Event",
                        related_name="line_quantities",
                        to="order.ShippingEvent",
                        on_delete=models.CASCADE,
                    ),
                ),
                (
                    "line",
                    models.ForeignKey(
                        verbose_name="Line",
                        related_name="shipping_event_quantities",
                        to="order.Line",
                        on_delete=models.CASCADE,
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "Shipping Event Quantities",
                "verbose_name": "Shipping Event Quantity",
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="ShippingEventType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(unique=True, max_length=255, verbose_name="Name"),
                ),
                (
                    "code",
                    oscar.models.fields.autoslugfield.AutoSlugField(
                        populate_from="name",
                        unique=True,
                        verbose_name="Code",
                        max_length=128,
                        editable=False,
                        blank=True,
                    ),
                ),  # NOQA
            ],
            options={
                "ordering": ("name",),
                "verbose_name_plural": "Shipping Event Types",
                "verbose_name": "Shipping Event Type",
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name="shippingeventquantity",
            unique_together={("event", "line")},
        ),
        migrations.AddField(
            model_name="shippingevent",
            name="event_type",
            field=models.ForeignKey(
                verbose_name="Event Type",
                to="order.ShippingEventType",
                on_delete=models.CASCADE,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="shippingevent",
            name="lines",
            field=models.ManyToManyField(
                related_name="shipping_events",
                verbose_name="Lines",
                to="order.Line",
                through="order.ShippingEventQuantity",
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="shippingevent",
            name="order",
            field=models.ForeignKey(
                verbose_name="Order",
                related_name="shipping_events",
                to="order.Order",
                on_delete=models.CASCADE,
            ),
            preserve_default=True,
        ),
        migrations.AlterUniqueTogether(
            name="paymenteventquantity",
            unique_together={("event", "line")},
        ),
        migrations.AddField(
            model_name="paymentevent",
            name="event_type",
            field=models.ForeignKey(
                verbose_name="Event Type",
                to="order.PaymentEventType",
                on_delete=models.CASCADE,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="paymentevent",
            name="lines",
            field=models.ManyToManyField(
                through="order.PaymentEventQuantity",
                verbose_name="Lines",
                to="order.Line",
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="paymentevent",
            name="order",
            field=models.ForeignKey(
                verbose_name="Order",
                related_name="payment_events",
                to="order.Order",
                on_delete=models.CASCADE,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="paymentevent",
            name="shipping_event",
            field=models.ForeignKey(
                related_name="payment_events",
                to="order.ShippingEvent",
                null=True,
                on_delete=models.CASCADE,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="order",
            name="shipping_address",
            field=models.ForeignKey(
                null=True,
                verbose_name="Shipping Address",
                on_delete=django.db.models.deletion.SET_NULL,
                to="order.ShippingAddress",
                blank=True,
            ),  # NOQA
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="order",
            name="site",
            field=models.ForeignKey(
                verbose_name="Site",
                on_delete=django.db.models.deletion.SET_NULL,
                to="sites.Site",
                null=True,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="order",
            name="user",
            field=models.ForeignKey(
                null=True,
                verbose_name="User",
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="orders",
                to=settings.AUTH_USER_MODEL,
                blank=True,
            ),  # NOQA
        ),
        migrations.AddField(
            model_name="lineprice",
            name="order",
            field=models.ForeignKey(
                verbose_name="Order",
                related_name="line_prices",
                to="order.Order",
                on_delete=models.CASCADE,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="line",
            name="order",
            field=models.ForeignKey(
                verbose_name="Order",
                related_name="lines",
                to="order.Order",
                on_delete=models.CASCADE,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="line",
            name="partner",
            field=models.ForeignKey(
                null=True,
                verbose_name="Partner",
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="order_lines",
                to="partner.Partner",
                blank=True,
            ),  # NOQA
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="line",
            name="product",
            field=models.ForeignKey(
                null=True,
                verbose_name="Product",
                on_delete=django.db.models.deletion.SET_NULL,
                to="catalogue.Product",
                blank=True,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="line",
            name="stockrecord",
            field=models.ForeignKey(
                null=True,
                verbose_name="Stock record",
                on_delete=django.db.models.deletion.SET_NULL,
                to="partner.StockRecord",
                blank=True,
            ),  # NOQA
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="communicationevent",
            name="order",
            field=models.ForeignKey(
                verbose_name="Order",
                related_name="communication_events",
                to="order.Order",
                on_delete=models.CASCADE,
            ),
            preserve_default=True,
        ),
    ]



import math
import random
import time

from oscar.apps.basket.models import Basket
from oscar.apps.order.utils import OrderCreator as BaseOrderCreator
from oscarapicheckout.mixins import OrderCreatorMixin


class OrderNumberGenerator:
    def order_number(self, basket: Basket) -> str:
        base_num = int(math.floor(time.time() / (3600 * 24)))
        basket_id = str(basket.id).zfill(3)[::-1]
        suffix = random.randrange(100, 999)
        number = "O" + str(base_num) + basket_id + str(suffix)
        return number


class OrderCreator(OrderCreatorMixin, BaseOrderCreator):
    pass



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("order", "0003_auto_20150113_1629"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="line",
            options={
                "ordering": ["pk"],
                "verbose_name": "Order Line",
                "verbose_name_plural": "Order Lines",
            },
        ),
    ]



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("order", "0002_auto_20141007_2032"),
    ]

    operations = [
        migrations.AlterField(
            model_name="order",
            name="date_placed",
            field=models.DateTimeField(db_index=True),
            preserve_default=True,
        ),
    ]



from django.http import HttpRequest
from oscar.core.loading import get_class

OrderPlacementMixin = get_class("checkout.mixins", "OrderPlacementMixin")


class OrderMessageSender(OrderPlacementMixin):
    def __init__(self, request: HttpRequest):
        self.request = request



# Generated by Django 1.10.5 on 2017-05-02 15:33

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("order", "0004_auto_20160111_1108"),
    ]

    operations = [
        migrations.AlterField(
            model_name="order",
            name="guest_email",
            field=models.EmailField(
                blank=True, max_length=254, verbose_name="Guest email address"
            ),
        ),
    ]



from collections.abc import Mapping
from decimal import Decimal
from typing import Any, TypedDict

from django.core.cache import cache
from django.utils.module_loading import import_string
from oscar.core.loading import get_model
from rest_framework import serializers

from . import settings as pkgsettings

Country = get_model("address", "Country")


class EmailAddressSerializer(serializers.Serializer[Any]):
    email = serializers.EmailField()


class EmailAddressCacheValue(TypedDict):
    email: str


class ShippingMethodSerializer(serializers.Serializer[Any]):
    code = serializers.CharField(max_length=128)
    name = serializers.CharField(max_length=128)
    price = serializers.DecimalField(decimal_places=2, max_digits=12)


class ShippingMethodCacheValue(TypedDict):
    code: str
    name: str
    price: Decimal


class BaseAddressCacheValue(TypedDict):
    title: str
    first_name: str
    last_name: str
    line1: str
    line2: str
    line3: str
    line4: str
    state: str
    postcode: str
    country: Country | None


class ShippingAddressCacheValue(BaseAddressCacheValue):
    phone_number: str
    notes: str


class BillingAddressCacheValue(BaseAddressCacheValue):
    pass


class AbstractCheckoutCache[
    ExternalData: Mapping[str, Any],
    StoredData: Mapping[str, Any],
]:
    serializer_class_path: str | None = None
    cache_timeout: int = 60 * 60 * 24  # 24 hours

    def __init__(self, basket_id: int, enable_validation: bool = False) -> None:
        self.basket_id = basket_id
        self.enable_validation = enable_validation

    @property
    def cache_key(self) -> str:
        return f"oscarapicheckout.cache.{self.__class__.__name__}.{self.basket_id}"

    def set(self, edata: ExternalData) -> None:
        sdata = self._transform_incoming_data(edata)
        cache.set(self.cache_key, sdata, self.cache_timeout)

    def get(self) -> ExternalData:
        data = cache.get(self.cache_key)
        return self._transform_outgoing_data(data)

    def invalidate(self) -> None:
        cache.delete(self.cache_key)

    def _transform_incoming_data(self, edata: ExternalData) -> StoredData:
        sdata: StoredData = edata  # type:ignore[assignment]
        if self.enable_validation:
            serializer = self._get_serializer(edata)
            if serializer:
                serializer.is_valid(raise_exception=True)
                sdata = serializer.validated_data
        return sdata

    def _transform_outgoing_data(self, sdata: StoredData) -> ExternalData:
        return sdata  # type:ignore[return-value]

    def _get_serializer(self, edata: ExternalData) -> serializers.Serializer[Any] | None:
        serializer_class = self._get_serializer_class()
        if serializer_class:
            return serializer_class(data=edata)
        return None

    def _get_serializer_class(self) -> type[serializers.Serializer[Any]] | None:
        if self.serializer_class_path:
            return import_string(  # type:ignore[no-any-return]
                self.serializer_class_path
            )
        return None


class AbstractCheckoutAddressCache[ExternalData: BaseAddressCacheValue](AbstractCheckoutCache[ExternalData, dict[str, Any]]):
    def _transform_incoming_data(self, edata: ExternalData) -> dict[str, Any]:
        sdata: dict[str, Any] = super()._transform_incoming_data(edata)
        if sdata and "country" in sdata and hasattr(sdata["country"], "pk"):
            sdata["country"] = sdata["country"].pk
        return sdata

    def _transform_outgoing_data(self, sdata: dict[str, Any]) -> ExternalData:
        edata: ExternalData = super()._transform_outgoing_data(sdata)
        if sdata and "country" in sdata and not hasattr(sdata["country"], "pk"):
            try:
                edata["country"] = Country.objects.get(pk=sdata["country"])
            except Country.DoesNotExist:
                edata["country"] = None
        return edata


class EmailAddressCache(
    AbstractCheckoutCache[
        EmailAddressCacheValue,
        dict[str, Any],
    ],
):
    serializer_class_path = pkgsettings.CHECKOUT_CACHE_SERIALIZERS.get("email_address", "oscarapicheckout.cache.EmailAddressSerializer")


class ShippingAddressCache(AbstractCheckoutAddressCache[ShippingAddressCacheValue]):
    serializer_class_path = pkgsettings.CHECKOUT_CACHE_SERIALIZERS.get("shipping_address", "oscarapi.serializers.checkout.ShippingAddressSerializer")


class BillingAddressCache(AbstractCheckoutAddressCache[BillingAddressCacheValue]):
    serializer_class_path = pkgsettings.CHECKOUT_CACHE_SERIALIZERS.get("billing_address", "oscarapi.serializers.checkout.BillingAddressSerializer")


class ShippingMethodCache(
    AbstractCheckoutCache[
        ShippingMethodCacheValue,
        dict[str, Any],
    ],
):
    serializer_class_path = pkgsettings.CHECKOUT_CACHE_SERIALIZERS.get("shipping_method", "oscarapicheckout.cache.ShippingMethodSerializer")



# Generated by Django 2.2.6 on 2020-02-19 09:16

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("order", "0008_auto_20190301_1035"),
    ]

    replaces = [
        ("order", "0008_surcharge"),
    ]

    operations = [
        migrations.CreateModel(
            name="Surcharge",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=128, verbose_name="Name")),
                (
                    "code",
                    models.CharField(max_length=128, verbose_name="Surcharge code"),
                ),
                (
                    "incl_tax",
                    models.DecimalField(
                        decimal_places=2,
                        default=0,
                        max_digits=12,
                        verbose_name="Surcharge (inc. tax)",
                    ),
                ),
                (
                    "excl_tax",
                    models.DecimalField(
                        decimal_places=2,
                        default=0,
                        max_digits=12,
                        verbose_name="Surcharge (excl. tax)",
                    ),
                ),
                (
                    "order",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="surcharges",
                        to="order.Order",
                        verbose_name="Order",
                    ),
                ),
            ],
            options={"abstract": False, "ordering": ["pk"]},
        ),
    ]



from collections.abc import Generator
from datetime import timedelta
from typing import Any, Literal, Protocol
import logging

from django.http import HttpRequest
from django.utils import timezone
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _
from oscar.core.loading import get_model
from rest_framework import serializers

from . import settings

Order = get_model("order", "Order")

logger = logging.getLogger(__name__)

type CheckoutData = dict[str, Any]


class FraudRule(Protocol):
    def validate(
        self,
        data: CheckoutData,
        recaptcha_score: float | None,
        request: HttpRequest | None,
    ) -> None: ...


def get_enabled_fraud_checks() -> Generator[FraudRule, None, None]:
    for config in settings.API_CHECKOUT_FRAUD_CHECKS:
        RuleClass: type[FraudRule] = import_string(config["rule"])
        rule = RuleClass(**config.get("kwargs", {}))
        yield rule


def run_enabled_fraud_checks(
    data: CheckoutData,
    recaptcha_score: float | None = None,
    request: HttpRequest | None = None,
) -> None:
    for rule in get_enabled_fraud_checks():
        rule.validate(data, recaptcha_score, request)
    return None


class AddressVelocity:
    """
    Built-in example of an order fraud check. Rejects new orders if there's been
    too many orders in a rolling time window with the same billing or shipping
    address.
    """

    period: timedelta
    threshold: int

    def __init__(
        self,
        period: timedelta | None = None,
        threshold: int = 10,
    ) -> None:
        self.period = period or timedelta(hours=24)
        self.threshold = threshold

    def validate(self, data: CheckoutData, *args: Any, **kwargs: Any) -> None:
        self._validate_addr("shipping_address", data)
        self._validate_addr("billing_address", data)

    def _validate_addr(
        self,
        addr_type: Literal["shipping_address", "billing_address"],
        data: CheckoutData,
    ) -> None:
        timeframe_start = timezone.now() - self.period
        filter_args = {
            "date_placed__gte": timeframe_start,
        }
        addr_data = data[addr_type]
        for addr_field in ["line1", "line2", "line3", "line4", "postcode"]:
            if addr_field in addr_data:
                filter_key = f"{addr_type}__{addr_field}__iexact"
                filter_args[filter_key] = addr_data[addr_field]

        if "country" in addr_data:
            filter_args[f"{addr_type}__country"] = addr_data["country"]

        address_use_count = Order.objects.filter(**filter_args).count()
        if address_use_count >= self.threshold:
            logger.info("Rejected order due to address velocity rules")
            raise serializers.ValidationError(_("Order rejected."))



# Generated by Django 4.2.6 on 2023-10-26 16:21

from django.db import migrations, models
import django.core.serializers.json


class Migration(migrations.Migration):
    dependencies = [
        ("order", "0010_auto_20210201_1735"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="surcharge",
            options={
                "ordering": ["pk"],
                "verbose_name": "Surcharge",
                "verbose_name_plural": "Surcharges",
            },
        ),
        migrations.AddField(
            model_name="line",
            name="tax_code",
            field=models.CharField(
                blank=True, max_length=64, null=True, verbose_name="VAT rate code"
            ),
        ),
        migrations.AddField(
            model_name="lineprice",
            name="tax_code",
            field=models.CharField(
                blank=True, max_length=64, null=True, verbose_name="VAT rate code"
            ),
        ),
        migrations.AddField(
            model_name="order",
            name="shipping_tax_code",
            field=models.CharField(
                blank=True,
                max_length=64,
                null=True,
                verbose_name="Shipping VAT rate code",
            ),
        ),
        migrations.AddField(
            model_name="surcharge",
            name="tax_code",
            field=models.CharField(
                blank=True, max_length=64, null=True, verbose_name="VAT rate code"
            ),
        ),
        migrations.AlterField(
            model_name="lineattribute",
            name="value",
            field=models.JSONField(
                encoder=django.core.serializers.json.DjangoJSONEncoder,
                verbose_name="Value",
            ),
        ),
    ]
